/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducer.h"
#include "gdamutil.h"
#include <gtk/gtksignal.h>
#include <stdio.h>

static void     gdam_producer_destroy       (GtkObject*    object);

static void     gdam_producer_real_seek     (GdamProducer* 	producer,
		          		     gdouble 		offset,
					     int 		whence);
static void     gdam_producer_real_stop     (GdamProducer*      producer);
static void     gdam_producer_real_set_position 
					    (GdamProducer*  producer,
	                                     gdouble        position);

	void		      (*source_ended) 
					  (GdamProducer* 	producer,
					   gdouble	 	end_position);
enum {
	PLAY,
	STOP,
	SEEK,
	SET_POSITION,
	UPDATE_LENGTH,
	SOURCE_ENDED,
	LAST_SIGNAL
};
static guint producer_signals[LAST_SIGNAL];
static GtkObjectClass* parent_class = 0;

static void gdam_producer_init(GdamProducer* producer)
{
	producer->stopped_position = 0.0;
}

static void gdam_producer_class_init(GdamProducerClass* klass)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	klass->seek = gdam_producer_real_seek;
	object_class->destroy = gdam_producer_destroy;
	klass->stop = gdam_producer_real_stop;
	klass->set_position = gdam_producer_real_set_position;
	producer_signals[PLAY] =
			gtk_signal_new(
				"play",
				GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamProducerClass, play),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE,
				1,
				GTK_TYPE_POINTER);
	producer_signals[STOP] =
			gtk_signal_new(
				"stop",
				GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamProducerClass, stop),
				gtk_marshal_DOUBLE__NONE,
				GTK_TYPE_DOUBLE,
				0);
	producer_signals[SEEK] =
			gtk_signal_new(
				"seek",
				GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamProducerClass, seek),
				gtk_marshal_NONE__DOUBLE_INT,
				GTK_TYPE_NONE,
				2,
				GTK_TYPE_DOUBLE,
				GTK_TYPE_INT);
	producer_signals[SET_POSITION] =
			gtk_signal_new(
				"set_position",
				GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamProducerClass, 
							set_position),
				gtk_marshal_NONE__DOUBLE,
				GTK_TYPE_NONE,
				1,
				GTK_TYPE_DOUBLE);
	producer_signals[UPDATE_LENGTH] =
			gtk_signal_new(
				"update_length",
				GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamProducerClass, 
							update_length),
				gtk_marshal_NONE__DOUBLE,
				GTK_TYPE_NONE,
				1,
				GTK_TYPE_DOUBLE);
	producer_signals[SOURCE_ENDED] =
			gtk_signal_new(
				"source_ended",
				GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamProducerClass, 
							source_ended),
				gtk_marshal_NONE__DOUBLE,
				GTK_TYPE_NONE,
				1,
				GTK_TYPE_DOUBLE);
	gtk_object_class_add_signals(object_class,
	                             producer_signals,
				     LAST_SIGNAL);
}

GtkType gdam_producer_get_type()
{
	static GtkType producer_type = 0;
	if (!producer_type) {
		static const GtkTypeInfo producer_info = {
			"GdamProducer",
			sizeof(GdamProducer),
			sizeof(GdamProducerClass),
			(GtkClassInitFunc) gdam_producer_class_init,
			(GtkObjectInitFunc) gdam_producer_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_MODEL);
		producer_type = gtk_type_unique(GDAM_TYPE_MODEL,
		                                &producer_info);
		
	}
	return producer_type;
}

void gdam_producer_play(GdamProducer* producer,
			GdamPlayInfo* play_info)
{
	g_return_if_fail(producer != 0);
	if (gtk_signal_n_emissions(GTK_OBJECT(producer),
				   producer_signals[PLAY]))
		return;
	if (play_info) {
		gtk_signal_emit(GTK_OBJECT(producer),
				producer_signals[PLAY],
				play_info);
	} else {
		GdamPlayInfo* p_info = g_new0(GdamPlayInfo, 1);
		gtk_signal_emit(GTK_OBJECT(producer),
				producer_signals[PLAY],
				&p_info);
	}
}

void gdam_producer_play_now(GdamProducer* producer)
{
	GdamPlayInfo* play_info = g_new0(GdamPlayInfo, 1);
	if (GDAM_MODEL(producer)->source_id != 0){
		gdam_debug("producer_play_now: have id, aborting.");
		return;
	}
	play_info->position = producer->stopped_position;
	gdam_debug("gdam_producer_play_now:  playing %p at %f sec",
			producer, play_info->position);
	gdam_producer_play(producer, play_info);
}

gdouble gdam_producer_stop(GdamProducer* producer)
{
	gdouble position;
	gtk_signal_emit(GTK_OBJECT(producer),
	                producer_signals[STOP],
			&position);
	return position;
}

void gdam_producer_seek(GdamProducer* producer, gdouble offset, int whence)
{
	gtk_signal_emit(GTK_OBJECT(producer),
	                producer_signals[SEEK],
			offset,
			whence);
}

static void 
gdam_producer_real_seek(GdamProducer* producer, gdouble offset, int whence)
{
	int source_id = GDAM_MODEL(producer)->source_id;
	GdamChannel* channel = GDAM_MODEL(producer)->channel;
	GdamTime tim = (GdamTime)(44100.0 * offset);
	GdamArg arg;
	GdamRequest* request;
	if (source_id == 0){
		double stopped_position = producer->stopped_position;
		switch (whence) {
			case 0:
				stopped_position = offset;
				break;
			case 1:
				stopped_position += offset;
				break;
			case 2:
				g_warning("cannot seek from end of file");
				break;
		}
		gdam_producer_set_position(producer, stopped_position);
		return;
	}
	switch (whence) {
		case 0: 
			arg.name = "position";
			arg.type = GDAM_ARG_TYPE_INT64;
			GDAM_ARG_INT64(&arg) = tim;
			request = gdam_request_set_args(
					source_id,
					1,
					&arg,
					gdam_do_nothing,
					gdam_do_nothing,
					NULL);
			gdam_channel_append_request(channel, request);
			break;
		case 1: 
			request = gdam_request_seek_relative(
					source_id,
					tim,
					gdam_do_nothing,
					gdam_do_nothing,
					NULL);
			gdam_channel_append_request(channel, request);
			break;
		case 2: 
			g_warning("cannot seek from end of file");
			break;
	}
}

static void     gdam_producer_destroy       (GtkObject*    object)
{
	GdamProducer* producer = (GdamProducer*)object;
	if (((GdamModel*)object)->source_id != 0)
		gdam_producer_stop(producer);
	(*parent_class->destroy)(object);
}

typedef struct _StopReturnValues StopReturnValues;
struct _StopReturnValues {
	GdamRequestGetTimeStruct source_time_return;
	GdamRequestGetIntStruct package_return;
	gboolean detach_source_return;
	GdamProducer* producer;
};

static void gdam_producer_get_stopped_position(gint64 pos, GdamProducer* prod)
{
	gdouble stopped_position = (gdouble) pos / 44100.0;
	gdam_debug("gdam_producer_get_stopped_position: %f seconds",
					stopped_position);
	gdam_producer_set_position(prod, stopped_position);
	gtk_object_unref(GTK_OBJECT(prod));
}
static void gdam_producer_failed_get_stopped_position(GdamProducer* prod)
{
	gdam_debug("failed getting stopped position, song had already ended?");
	gtk_object_unref(GTK_OBJECT(prod));
}

/* TODO: if making_request, register for that source to be removed on completion
 */
static void  gdam_producer_real_stop        (GdamProducer*    producer)
{
	GdamModel* model = GDAM_MODEL(producer);
	GdamRequestBuilder* builder;
	StopReturnValues* ret_values = g_new(StopReturnValues, 1);
	if (model->source_id == 0){
		gdam_debug("real_stop: source id zero.");
		return;
	}

	builder = gdam_request_builder_new();

	/* This reference will be undone be stop_failed_callback
	 * or stop_succeeded_callback, */
	gtk_object_ref(GTK_OBJECT(producer));

	ret_values->producer = producer;
	gdam_request_builder_append_generic(
			builder,
			gdam_request_source_time(
				   model->source_id,
				   0, 
				   (GdamRequestGotTimeFunc) 
				     gdam_producer_get_stopped_position,
				   (GHookFunc)
				     gdam_producer_failed_get_stopped_position,
				   producer
			)
	);
	gdam_request_builder_append_generic(
			builder,
			gdam_request_detach_source(
				   model->source_id,
				   NULL,
				   NULL,
				   NULL
			)
	);
	gdam_model_new_child_request(
				model->parent,
				model,
				NULL,
				0,
				builder
	);
	gdam_model_set_source_id(model, 0);
	return;
}

static void gdam_producer_real_set_position (GdamProducer*  producer,
	                                     gdouble        position)
{
	producer->stopped_position = position;
}

void       gdam_producer_set_position (GdamProducer*  producer,
	                               gdouble        position)
{
	if (gtk_signal_n_emissions(GTK_OBJECT(producer),
				   producer_signals[SET_POSITION]))
		return;
        gtk_signal_emit( GTK_OBJECT(producer),
			 producer_signals[SET_POSITION],
			 position);
}

void       gdam_producer_update_length (GdamProducer*  producer,
	                                gdouble        length)
{
	if (gtk_signal_n_emissions(GTK_OBJECT(producer),
				   producer_signals[UPDATE_LENGTH]))
		return;
        gtk_signal_emit( GTK_OBJECT(producer),
			 producer_signals[UPDATE_LENGTH],
			 length);
}

void       gdam_producer_source_ended (GdamProducer*  producer,
	                               gdouble        position)
{
	if (gtk_signal_n_emissions(GTK_OBJECT(producer),
				   producer_signals[SOURCE_ENDED]))
		return;
        gtk_signal_emit( GTK_OBJECT(producer),
			 producer_signals[SOURCE_ENDED],
			 position);
}
