/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamfirfilter.h"
#include <gdam/gdamsample.h>

static gboolean gdam_fir_filter_refresh_echo_array(GdamEchoBase* echo_base)
{
	GdamFirFilter* fir_filter = GDAM_FIR_FILTER(echo_base);
	int* echoes = g_new(int, fir_filter->num_elements * 2);
	int i, num_echoes = fir_filter->num_elements;
	GdamFirFilterElement* elements = fir_filter->elements;
	gdouble vol_scale = fir_filter->gain_scale * GdamFixed_ONE;

	echo_base->num_echoes = num_echoes;

	if (fir_filter->normalize) {
		gdouble total = 0.0;
		for (i = 0; i < num_echoes; i++) {
			gdouble vol = elements[i].volume;
			total += ABS(vol);
		}
		if (total > 0.0)
			vol_scale /= total;
	}

	for (i = 0; i < num_echoes; i++) {
		gdouble volume = elements[i].volume;
		gdouble delay = elements[i].offset;
		g_message("refresh_echo_array: echo %d, delay %d, vol %d", i, (int)delay, (int)(vol_scale * volume));
		delay *= fir_filter->grid_unit;
		echoes[2*i+0] = (int) (volume * vol_scale);
		echoes[2*i+1] = (int) delay;
	}

	g_free(echo_base->echoes);
	echo_base->echoes = echoes;

	return TRUE;
}

#define IS_POWER_OF_TWO(i)		(((i) & ((i)-1)) == 0)
void gdam_fir_filter_add(GdamFirFilter*        fir_filter,
                         gdouble               gain,
			 gdouble               delay)
{
	int num_elements = fir_filter->num_elements;
	if (num_elements == 0) {
		fir_filter->elements = g_new(GdamFirFilterElement, 32);
	} else if (num_elements >= 32 && IS_POWER_OF_TWO(num_elements)) {
		fir_filter->elements = g_renew(GdamFirFilterElement,
		                         fir_filter->elements,
		                         num_elements * 2);
	}

	fir_filter->elements[num_elements].volume = gain;
	fir_filter->elements[num_elements].offset = delay;
	fir_filter->num_elements = ++num_elements;
	gdam_echo_base_refresh_echo_array(GDAM_ECHO_BASE(fir_filter));
}

static void gdam_fir_filter_init(GdamFirFilter* fir_filter)
{
	fir_filter->grid_unit = 1.0;
	fir_filter->num_elements = 0;
	fir_filter->elements = NULL;
	fir_filter->gain_scale = 1.0;
	gdam_fir_filter_add(fir_filter, 1.0, 0.0);
}

static void gdam_fir_filter_class_init(GdamFirFilterClass* fir_filter_class);

GtkType gdam_fir_filter_get_type()
{
	static GtkType fir_filter_type = 0;
	if (!fir_filter_type) {
		static const GtkTypeInfo fir_filter_info = {
			"GdamFirFilter",
			sizeof(GdamFirFilter),
			sizeof(GdamFirFilterClass),
			(GtkClassInitFunc) gdam_fir_filter_class_init,
			(GtkObjectInitFunc) gdam_fir_filter_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_ECHO_BASE;
		fir_filter_type = gtk_type_unique(parent,
		                                 &fir_filter_info);
	}
	return fir_filter_type;
}

static void gdam_fir_filter_arg_set_normalized(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	firfilter->normalize = GDAM_ARG_INT(arg);
	gdam_echo_base_refresh_echo_array(GDAM_ECHO_BASE(firfilter));
}

static void gdam_fir_filter_arg_get_normalized(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	GDAM_ARG_INT(arg) = firfilter->normalize;
}

static void gdam_fir_filter_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	GdamBeatInfo* beat_info = &GDAM_MODEL(firfilter)->beat_info;
	gdouble bpm = gdam_beat_info_get_bpm(beat_info, 0);
	GdamArg new_arg;
	new_arg.name = "grid_unit";
	new_arg.type = GDAM_ARG_TYPE_DOUBLE;
	firfilter->beat_factor = GDAM_ARG_DOUBLE(arg);
	if (bpm < .001)
		return;
	GDAM_ARG_DOUBLE(&new_arg) = 60.0 * 44100.0 / bpm;
	gdam_arg_set_object_arg(GTK_OBJECT(firfilter), &new_arg);
	gdam_echo_base_refresh_echo_array(GDAM_ECHO_BASE(firfilter));
}

static void gdam_fir_filter_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	GDAM_ARG_DOUBLE(arg) = firfilter->beat_factor;
}

static void gdam_fir_filter_arg_set_grid_unit(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	firfilter->grid_unit = GDAM_ARG_DOUBLE(arg);
	gdam_echo_base_refresh_echo_array(GDAM_ECHO_BASE(firfilter));
}

static void gdam_fir_filter_arg_get_grid_unit(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	GDAM_ARG_DOUBLE(arg) = firfilter->grid_unit;
}

static void gdam_fir_filter_arg_set_gain_scale(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	firfilter->gain_scale = GDAM_ARG_DOUBLE(arg);
	gdam_echo_base_refresh_echo_array(GDAM_ECHO_BASE(firfilter));
}

static void gdam_fir_filter_arg_get_gain_scale(GtkObject* object, GdamArg* arg)
{
	GdamFirFilter* firfilter = GDAM_FIR_FILTER(object);
	GDAM_ARG_DOUBLE(arg) = firfilter->gain_scale;
}

static void gdam_fir_filter_class_init(GdamFirFilterClass* fir_filter_class)
{
	GdamEchoBaseClass* echo_base_class;
	GtkObjectClass* object_class;
	object_class = GTK_OBJECT_CLASS(fir_filter_class);
	echo_base_class = GDAM_ECHO_BASE_CLASS(fir_filter_class);
	echo_base_class->refresh_echo_array
		= gdam_fir_filter_refresh_echo_array;
	gdam_arg_register(object_class,
			  "normalized",
			  GDAM_ARG_TYPE_INT,
			  gdam_fir_filter_arg_set_normalized,
			  gdam_fir_filter_arg_get_normalized,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
			  "gain_scale",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_fir_filter_arg_set_gain_scale,
			  gdam_fir_filter_arg_get_gain_scale,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
			  "grid_unit",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_fir_filter_arg_set_grid_unit,
			  gdam_fir_filter_arg_get_grid_unit,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
			  "beat_factor",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_fir_filter_arg_set_beat_factor,
			  gdam_fir_filter_arg_get_beat_factor,
			  GDAM_ARG_VITAL);
}
