/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamcachebuffer.h"
#include <gdam/gdam-protocol.h>

	
/*
 * For displaying a GdamCacheBuffer on the client side.
 */

typedef void (*CopyFunc)(int* out, GdamSample* in);
static void copy_left_sample(int* out, GdamSample* in)
{
	*out = in->s[0];
}
static void copy_right_sample(int* out, GdamSample* in)
{
	*out = in->s[1];
}
static void copy_average_sample(int* out, GdamSample* in)
{
	*out = ((int)in->s[0] + in->s[1]) / 2;
}
static void copy_stereo_sample(int* out, GdamSample* in)
{
	*out++ = in->s[0];
	*out++ = in->s[1];
}

gint*     gdam_cache_buffer_profile(GdamCacheBuffer*        buffer,
				    int                     mode,
                                    int                     sample_start,
				    int                     sample_length,
				    int                     num_outputs)
{
	int prepad_zeroes = 0, postpad_zeroes = 0;
	int orig_num_outputs = num_outputs;
	int channels = (mode == GDAM_PROTOCOL_PROFILE_STEREO ? 2 : 1);
	CopyFunc copy_func = NULL;
	gint* outputs;
	int* at;
	int i;
	int num_ints;

	/* Does it overrun the sample's start? */
	if (sample_start < 0) {
		prepad_zeroes = (-sample_start) * num_outputs / sample_length;
		sample_length += sample_start;
		sample_start = 0;
		num_outputs -= prepad_zeroes;
	}

	/* Allocate the output area... */
	num_ints = channels * orig_num_outputs;
	outputs = g_new(gint, num_ints);

	if (sample_length <= 0 || prepad_zeroes < 0) {
		bzero(outputs, sizeof(gint) * num_ints);
		return outputs;
	}

	/* Does it overrun the sample's end? */
	if (sample_start + sample_length > buffer->buffer_length) {
		int cut = sample_start + sample_length - buffer->buffer_length;
		postpad_zeroes = (gint64) cut * num_outputs / sample_length;
		num_outputs -= postpad_zeroes;
		sample_length -= cut;
	}

	/* If they are hopelessly out of range, fill the buffer
	 * with silence and return. */
	if (num_outputs <= 0) {
		bzero(outputs, sizeof(gint) * num_ints);
		return outputs;
	}

	/* Zero any padding we had to insert. */
	bzero(outputs, channels * prepad_zeroes * sizeof(gint));

	at = outputs + channels * prepad_zeroes;
	switch (mode) {
		case GDAM_PROTOCOL_PROFILE_LEFT:
			copy_func = copy_left_sample;
			break;
		case GDAM_PROTOCOL_PROFILE_RIGHT:
			copy_func = copy_right_sample;
			break;
		case GDAM_PROTOCOL_PROFILE_STEREO:
			copy_func = copy_stereo_sample;
			break;
		case GDAM_PROTOCOL_PROFILE_AVERAGE:
			copy_func = copy_average_sample;
			break;
		default: {
			g_warning("cache_profile mode not known: %d", mode);
			return NULL;
		}
	}
	for (i = 0; i < num_outputs; i++) {
		gint64 index = i;
		index *= sample_length;
		index /= num_outputs;
		index += sample_start;
		(*copy_func)(at, buffer->buffer + index);
		at += channels;
	}
	/* Zero any padding we had to insert. */
	bzero(at, channels * postpad_zeroes * sizeof(gint));
	return outputs;
}
