/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_SOURCE_H__
#define __GDAM_SOURCE_H__

typedef struct _GdamSource GdamSource;
typedef struct _GdamSourceClass GdamSourceClass;
typedef struct _GdamLatencyInfo GdamLatencyInfo;

#include <gtk/gtkobject.h>
#include "gdamsink.h"
#include <gdam/gdambuffer.h>
#include <gdam/gdamsample.h>
#include <gdam/gdamarg.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


GtkType gdam_source_get_type();
#define GDAM_TYPE_SOURCE              (gdam_source_get_type ())
#define GDAM_SOURCE(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_SOURCE, GdamSource))
#define GDAM_SOURCE_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_SOURCE, GdamSourceClass))
#define GDAM_IS_SOURCE(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_SOURCE))
#define GDAM_IS_SOURCE_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_SOURCE))

guint gdam_source_get_ended_event();
#define GDAM_SOURCE_EVENT_ENDED		(gdam_source_get_ended_event())

guint gdam_source_event_timer();
#define GDAM_SOURCE_EVENT_TIMER         (gdam_source_event_timer())


/* Flags for `write_to_binary'. */
#define GDAM_SOURCE_WRITE_NO_RECURSE	1

/*
 * Latency information:
 */
struct _GdamLatencyInfo {
	int			latency;
	GdamFixed		resample;
};

struct _GdamSource {
	GtkObject		object;
	
	GdamSource*		parent;
	GdamSink*		device;
	int			source_id;

	unsigned		ended : 1,
				paused : 1;
};

struct _GdamSourceClass {
	GtkObjectClass parent_class;
	GdamBuffer*       (*interpret_command) (GdamSource*      source,
	                                        GdamBuffer*      command,
						gboolean*        was_success);
	void              (*get_latency_info)  (GdamSource*      source,
	                                        GdamSource*      subsource,
	                                        GdamLatencyInfo* latency);

        void              (*done_constructing) (GdamSource*      source);

	/* If this is NULL, then the source is always ready-to-read. */
	gboolean          (*is_ready_to_read)  (GdamSource*      source);

	/* Get the position... only used for user queries; do not
	 * rely on it internally!!! */
	GdamTime          (*get_position)      (GdamSource*      source);

	/* Get the subsources of a source.
	 * The caller should expect an array in *children
	 * which must be freed by the caller. */
	void              (*get_subsources)    (GdamSource*      source,
	                                        int*             num_children,
						GdamSource***    children);

	/* Return values:
	 *      positive: the number of bytes written into buffer,
	 *      0 normal end of stream,
	 *      -1 error end of stream.
	 */
	int		   (*read)             (GdamSource*      input,
	                                        GdamSample*      samples,
						int              max_read);
	
#if 0
	int                (*read_full)        (GdamSource*      input,
	                                        GdamSampleInfo*  sample_format,
						void*            sample_data,
						int              max_read);
#endif

	/* Seek a sound source. */
	gboolean	   (*seek)             (GdamSource*      input,
	                                        GdamTime         offset,
						int              whence);

	/* Inform the source that some data was lost. */
	void               (*miss_data)        (GdamSource*      input,
	                                        int              samples_lost);

	/* For container sources, remove that child. */
	gboolean           (*detach_child)     (GdamSource*      parent, 
	                                        GdamSource*      child);

	/* Called by gdam_source_read (or whoever else uses the
	 * read virtual) to indicate the source has EOF'd. */
	void               (*ended)            (GdamSource*      source);
};

#ifdef __cplusplus
}
#endif /* __cplusplus */


GdamSource*        gdam_source_new_generic_v(const char* type,
                                             int         num_args,
					     GdamArg*    args);
GdamSource*        gdam_source_new_generic  (const char* type,
                                             int         num_args,
					     ...);
void               gdam_source_set_gdam_arg (GdamSource* source, 
                                             GdamArg*    arg);
void               gdam_source_get_gdam_arg (GdamSource* source, 
                                             GdamArg*    arg);


gboolean           gdam_source_detach_child (GdamSource* parent,
                                             GdamSource* child);

/* Get a source class from a well-known (by the gdam protocol)
 * type number.
 * Used by `gdamdispatch' to get at our idea of things from
 * the well-known protocol.
 */
GtkType            gdam_source_type_to_gtk_type(int gdam_protocol_type);

/* Detach it from whatever source it may be feeding into.
 *
 * FIXME (and all dependant code): increments the source's ref count... */
void               gdam_source_orphan          (GdamSource*      source);

/* Compute how far to seek a source to sync it. */
gint               gdam_source_get_sync_seek   (GdamSource*      mobile,
                                                GdamSource*      fixed,

						/* We cheat and sneak this 
						 * value in b32... */
						GdamTime         mobile_period,
						/* Just in samples... */
						GdamTime         mobile_offset,

						/* We cheat again... */
						GdamTime         fixed_period,
						/* Just in samples... */
						GdamTime         fixed_offset,
						
						/* Something to use for
						 * latency if we aren't added
						 * yet... */
						GdamSource*      mobile_lat);


gboolean           gdam_source_is_ancestor      (GdamSource*     child,
                                                 GdamSource*
						           potential_ancestor);
gboolean           gdam_source_is_ready_to_read (GdamSource*     source);
int                gdam_source_read             (GdamSource*     input,
                                                 GdamSample*     samples,
					         int             max_read);
gboolean           gdam_source_seek             (GdamSource*     source,
                                                 GdamTime        offset,
						 int             whence);
void               gdam_source_miss_data        (GdamSource*     input,
                                                 int             samples_lost);
GdamBuffer*        gdam_source_interpret_command(GdamSource*     source,
	                                         GdamBuffer*     command,
						 gboolean*       was_success);
GdamTime           gdam_source_get_position     (GdamSource*     source);
void               gdam_source_find_total_latency(GdamSource*    source,
			                         GdamLatencyInfo* total);

/* The caller must free the array returned in *subs_out. */
void               gdam_source_get_subsources   (GdamSource*     source, 
                                                 int*            num_subs_out,
						 GdamSource***   subs_out);

/* Functions to globally freeze and unfreeze done_constructing callbacks.
 * This is used inside package request handlers to prevent multiple
 * tweak from creating too many mpg123s. */
void gdam_source_defer_done_constructing();
void gdam_source_undefer_done_constructing();

#endif
